# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Accretive namespaces.

    Provides a namespace type that can grow but never shrink. Once an attribute is
    set, it cannot be modified or removed. This provides a simple way to create
    objects with named attributes that become immutable after assignment.

    The namespace implementation is modeled after :py:class:`types.SimpleNamespace`
    but adds accretive behavior. Like :py:class:`types.SimpleNamespace`, it
    provides a simple ``__repr__`` which lists all attributes.

    >>> from accretive import Namespace
    >>> ns = Namespace( apples = 12, bananas = 6 )
    >>> ns.cherries = 42  # Add new attribute
    >>> ns.apples = 14    # Attempt modification
    Traceback (most recent call last):
        ...
    accretive.exceptions.AttributeImmutability: Could not assign or delete existing attribute 'apples'.
    >>> del ns.bananas    # Attempt deletion
    Traceback (most recent call last):
        ...
    accretive.exceptions.AttributeImmutability: Could not assign or delete existing attribute 'bananas'.
''' # noqa: E501


from . import __
from . import iclasses as _iclasses


class Namespace( # noqa: PLW1641
    metaclass = _iclasses.Class,
    instances_assigner_core = _iclasses.assign_attribute_if_absent_mutable,
):
    # TODO: Dynadoc fragments.
    ''' Accretive namespaces. '''

    __slots__ = ( '__dict__', )

    def __init__(
        self,
        *iterables: __.DictionaryPositionalArgument[ __.H, __.V ],
        **attributes: __.DictionaryNominativeArgument[ __.V ],
    ) -> None:
        super( ).__init__( )
        super( ).__getattribute__( '__dict__' ).update(
            __.AccretiveDictionary( *iterables, **attributes ) )

    def __repr__( self ) -> str:
        attributes = ', '.join( tuple(
            f"{key} = {value!r}" for key, value
            in getattr( self, '__dict__', { } ).items( ) ) )
        fqname = __.ccutils.qualify_class_name( type( self ) )
        if not attributes: return f"{fqname}( )"
        return f"{fqname}( {attributes} )"

    def __eq__( self, other: __.typx.Any ) -> __.ComparisonResult:
        if isinstance( other, ( Namespace, __.types.SimpleNamespace ) ):
            return self.__dict__ == other.__dict__
        return NotImplemented

    def __ne__( self, other: __.typx.Any ) -> __.ComparisonResult:
        if isinstance( other, ( Namespace, __.types.SimpleNamespace ) ):
            return self.__dict__ != other.__dict__
        return NotImplemented



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Accretive data structures.

    Data structures which can grow but never shrink - once values are set, they
    become immutable. This behavior is useful for configuration registries,
    plugin systems, and other scenarios requiring grow-only collections with
    immutability guarantees.
'''


from . import __
# --- BEGIN: Injected by Copier ---
from . import exceptions
# --- END: Injected by Copier ---

from .classes import *
from .dictionaries import *
from .modules import *
from .namespaces import *


__version__: __.typx.Annotated[ str, __.ddoc.Visibilities.Reveal ]
__version__ = '4.0'


__.ccstd.dynadoc.assign_module_docstring(
    __name__,
    introspection = __.dynadoc_introspection_control_on_package,
    table = __.fragments )
reclassify_modules( __name__, recursive = True )



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Common names and type aliases. '''

# ruff: noqa: F401


from . import imports as __


from classcore.standard.nomina import (
                            AssignerLigation,
                            AttributesNamer,
                            BehaviorExclusionNamesOmni,
                            BehaviorExclusionRegexes,
                            BehaviorExclusionPredicates,
                            BehaviorExclusionVerifiersOmni,
    DecorationPreparers as  ClassDecorationPreparers,
                            DynadocConfiguration,
                            ErrorClassProvider,
                            concealment_label,
                            immutability_label,
)


H = __.typx.TypeVar( 'H', bound = __.cabc.Hashable ) # Hash Key
T = __.typx.TypeVar( 'T', bound = type ) # Class
U = __.typx.TypeVar( 'U' )  # Class
V = __.typx.TypeVar( 'V' ) # Value


ComparisonResult: __.typx.TypeAlias = bool | __.types.NotImplementedType
NominativeArguments: __.typx.TypeAlias = __.cabc.Mapping[ str, __.typx.Any ]
PositionalArguments: __.typx.TypeAlias = __.cabc.Sequence[ __.typx.Any ]

# TODO: Import ClassDecorator aliases from 'classcore' once documentation
#       fragments have been removed from them.
ClassDecorator: __.typx.TypeAlias = (
    __.cabc.Callable[ [ type[ U ] ], type[ U ] ] )
ClassDecorators: __.typx.TypeAlias = (
    __.cabc.Sequence[ ClassDecorator[ U ] ] )
ClassDecoratorFactory: __.typx.TypeAlias = (
    __.cabc.Callable[ ..., ClassDecorator[ U ] ] )
ModuleReclassifier: __.typx.TypeAlias = __.cabc.Callable[
    [ __.cabc.Mapping[ str, __.typx.Any ] ], None ]

DictionaryNominativeArgument: __.typx.TypeAlias = __.typx.Annotated[
    V,
    __.ddoc.Doc(
        'Zero or more keyword arguments from which to initialize '
        'dictionary data.' ),
]
DictionaryPositionalArgument: __.typx.TypeAlias = __.typx.Annotated[
    __.cabc.Mapping[ H, V ] | __.cabc.Iterable[ tuple[ H, V ] ],
    __.ddoc.Doc(
        'Zero or more iterables from which to initialize dictionary data. '
        'Each iterable must be dictionary or sequence of key-value pairs. '
        'Duplicate keys will result in an error.' ),
]
DictionaryProducer: __.typx.TypeAlias = __.typx.Annotated[
    __.cabc.Callable[ [ ], V ],
    __.ddoc.Doc(
        'Callable which produces values for absent dictionary entries.' ),
]
DictionaryValidator: __.typx.TypeAlias = __.typx.Annotated[
    __.cabc.Callable[ [ H, V ], bool ],
    __.ddoc.Doc(
        'Callable which validates entries before addition to dictionary.' ),
]


package_name = __name__.split( '.', maxsplit = 1 )[ 0 ]


def calculate_attrname( level: str, core: str ) -> str:
    return f"_{package_name}_{level}_{core}_"


# TODO: Import 'is_public_identifier' from 'classcore'.
def is_public_identifier( name: str ) -> bool:
    ''' Is Python identifier public? '''
    return not name.startswith( '_' )



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Accretive classes. '''

# ruff: noqa: F811


from . import __
from . import iclasses as _iclasses


mutables_default = _iclasses.mutables_default
visibles_default = _iclasses.visibles_default


_class_factory = __.funct.partial(
    __.ccstd.class_factory,
    assigner_core = _iclasses.assign_attribute_if_absent_mutable,
    attributes_namer = __.calculate_attrname,
    dynadoc_configuration = _iclasses.dynadoc_configuration,
    error_class_provider = _iclasses.provide_error_class )


@_class_factory( )
class Class( type ):
    ''' Metaclass for standard classes. '''

    _dynadoc_fragments_ = (
        'cfc class conceal', 'cfc class accrete', 'cfc dynadoc',
        'cfc instance conceal', 'cfc instance protect' )

    def __new__( # Typechecker stub.
        clscls: type[ __.T ],
        name: str,
        bases: tuple[ type, ... ],
        namespace: dict[ str, __.typx.Any ], *,
        decorators: __.ClassDecorators[ __.T ] = ( ),
        **arguments: __.typx.Unpack[ __.ccstd.ClassFactoryExtraArguments ],
    ) -> __.T:
        return super( ).__new__( clscls, name, bases, namespace )


@_class_factory( )
@__.typx.dataclass_transform( frozen_default = True, kw_only_default = True )
class Dataclass( type ):
    ''' Metaclass for standard dataclasses. '''

    _dynadoc_fragments_ = (
        'cfc produce dataclass',
        'cfc class conceal', 'cfc class accrete', 'cfc dynadoc',
        'cfc instance conceal', 'cfc instance protect' )

    def __new__( # Typechecker stub.
        clscls: type[ __.T ],
        name: str,
        bases: tuple[ type, ... ],
        namespace: dict[ str, __.typx.Any ], *,
        decorators: __.ClassDecorators[ __.T ] = ( ),
        **arguments: __.typx.Unpack[ __.ccstd.ClassFactoryExtraArguments ],
    ) -> __.T:
        return super( ).__new__( clscls, name, bases, namespace )


@_class_factory( )
@__.typx.dataclass_transform( kw_only_default = True )
class DataclassMutable( type ):
    ''' Metaclass for dataclasses with mutable instance attributes. '''

    _dynadoc_fragments_ = (
        'cfc produce dataclass',
        'cfc class conceal', 'cfc class accrete', 'cfc dynadoc',
        'cfc instance conceal' )

    def __new__( # Typechecker stub.
        clscls: type[ __.T ],
        name: str,
        bases: tuple[ type, ... ],
        namespace: dict[ str, __.typx.Any ], *,
        decorators: __.ClassDecorators[ __.T ] = ( ),
        **arguments: __.typx.Unpack[ __.ccstd.ClassFactoryExtraArguments ],
    ) -> __.T: # pragma: no cover
        return super( ).__new__( clscls, name, bases, namespace )


@_class_factory( )
class AbstractBaseClass( __.abc.ABCMeta ):
    ''' Metaclass for standard abstract base classes. '''

    _dynadoc_fragments_ = (
        'cfc produce abstract base class',
        'cfc class conceal', 'cfc class accrete', 'cfc dynadoc',
        'cfc instance conceal', 'cfc instance protect' )

    def __new__( # Typechecker stub.
        clscls: type[ __.T ],
        name: str,
        bases: tuple[ type, ... ],
        namespace: dict[ str, __.typx.Any ], *,
        decorators: __.ClassDecorators[ __.T ] = ( ),
        **arguments: __.typx.Unpack[ __.ccstd.ClassFactoryExtraArguments ],
    ) -> __.T:
        return super( ).__new__( clscls, name, bases, namespace )


@_class_factory( )
class ProtocolClass( type( __.typx.Protocol ) ):
    ''' Metaclass for standard protocol classes. '''

    _dynadoc_fragments_ = (
        'cfc produce protocol class',
        'cfc class conceal', 'cfc class accrete', 'cfc dynadoc',
        'cfc instance conceal', 'cfc instance protect' )

    def __new__( # Typechecker stub.
        clscls: type[ __.T ],
        name: str,
        bases: tuple[ type, ... ],
        namespace: dict[ str, __.typx.Any ], *,
        decorators: __.ClassDecorators[ __.T ] = ( ),
        **arguments: __.typx.Unpack[ __.ccstd.ClassFactoryExtraArguments ],
    ) -> __.T:
        return super( ).__new__( clscls, name, bases, namespace )


@_class_factory( )
@__.typx.dataclass_transform( frozen_default = True, kw_only_default = True )
class ProtocolDataclass( type( __.typx.Protocol ) ):
    ''' Metaclass for standard protocol dataclasses. '''

    _dynadoc_fragments_ = (
        'cfc produce protocol class', 'cfc produce dataclass',
        'cfc class conceal', 'cfc class accrete', 'cfc dynadoc',
        'cfc instance conceal', 'cfc instance protect' )

    def __new__( # Typechecker stub.
        clscls: type[ __.T ],
        name: str,
        bases: tuple[ type, ... ],
        namespace: dict[ str, __.typx.Any ], *,
        decorators: __.ClassDecorators[ __.T ] = ( ),
        **arguments: __.typx.Unpack[ __.ccstd.ClassFactoryExtraArguments ],
    ) -> __.T: # pragma: no cover
        return super( ).__new__( clscls, name, bases, namespace )


@_class_factory( )
@__.typx.dataclass_transform( kw_only_default = True )
class ProtocolDataclassMutable( type( __.typx.Protocol ) ):
    ''' Metaclass for protocol dataclasses with mutable instance attributes.
    '''

    _dynadoc_fragments_ = (
        'cfc produce protocol class', 'cfc produce dataclass',
        'cfc class conceal', 'cfc class accrete', 'cfc dynadoc',
        'cfc instance conceal' )

    def __new__( # Typechecker stub.
        clscls: type[ __.T ],
        name: str,
        bases: tuple[ type, ... ],
        namespace: dict[ str, __.typx.Any ], *,
        decorators: __.ClassDecorators[ __.T ] = ( ),
        **arguments: __.typx.Unpack[ __.ccstd.ClassFactoryExtraArguments ],
    ) -> __.T: # pragma: no cover
        return super( ).__new__( clscls, name, bases, namespace )


class Object(
    metaclass = _iclasses.Class,
    instances_assigner_core = _iclasses.assign_attribute_if_absent_mutable,
):
    ''' Standard base class. '''

    _dynadoc_fragments_ = (
        'class concealment', 'class protection', 'class dynadoc',
        'class instance conceal', 'class instance accrete' )


class ObjectMutable(
    metaclass = _iclasses.Class,
    instances_assigner_core = _iclasses.assign_attribute_if_absent_mutable,
    instances_mutables = '*',
):
    ''' Base class with mutable instance attributes. '''

    _dynadoc_fragments_ = (
        'class concealment', 'class protection', 'class dynadoc',
        'class instance conceal' )


class DataclassObject(
    metaclass = _iclasses.Dataclass,
    instances_assigner_core = _iclasses.assign_attribute_if_absent_mutable,
):
    ''' Standard base dataclass. '''

    _dynadoc_fragments_ = (
        'dataclass',
        'class concealment', 'class protection', 'class dynadoc',
        'class instance conceal', 'class instance accrete' )


class DataclassObjectMutable(
    metaclass = _iclasses.DataclassMutable,
    instances_assigner_core = _iclasses.assign_attribute_if_absent_mutable,
):
    ''' Base dataclass with mutable instance attributes. '''

    _dynadoc_fragments_ = (
        'dataclass',
        'class concealment', 'class protection', 'class dynadoc',
        'class instance conceal' )


class Protocol(
    __.typx.Protocol,
    metaclass = _iclasses.ProtocolClass,
    instances_assigner_core = _iclasses.assign_attribute_if_absent_mutable,
):
    ''' Standard base protocol class. '''

    _dynadoc_fragments_ = (
        'protocol class',
        'class concealment', 'class protection', 'class dynadoc',
        'class instance conceal', 'class instance accrete' )


class ProtocolMutable(
    __.typx.Protocol,
    metaclass = _iclasses.ProtocolClass,
    instances_assigner_core = _iclasses.assign_attribute_if_absent_mutable,
    instances_mutables = '*',
):
    ''' Base protocol class with mutable instance attributes. '''

    _dynadoc_fragments_ = (
        'protocol class',
        'class concealment', 'class protection', 'class dynadoc',
        'class instance conceal' )


class DataclassProtocol(
    __.typx.Protocol,
    metaclass = _iclasses.ProtocolDataclass,
    instances_assigner_core = _iclasses.assign_attribute_if_absent_mutable,
):
    ''' Standard base protocol dataclass. '''

    _dynadoc_fragments_ = (
        'dataclass', 'protocol class',
        'class concealment', 'class protection', 'class dynadoc',
        'class instance conceal', 'class instance accrete' )


class DataclassProtocolMutable(
    __.typx.Protocol,
    metaclass = _iclasses.ProtocolDataclassMutable,
    instances_assigner_core = _iclasses.assign_attribute_if_absent_mutable,
):
    ''' Base protocol dataclass with mutable instance attributes. '''

    _dynadoc_fragments_ = (
        'dataclass', 'protocol class',
        'class concealment', 'class protection', 'class dynadoc',
        'class instance conceal' )


@__.typx.overload
def dataclass_with_standard_behaviors( # pragma: no cover
    cls: type[ __.U ], /, *,
    decorators: __.ClassDecorators[ __.U ] = ( ),
    mutables: __.BehaviorExclusionVerifiersOmni = mutables_default,
    visibles: __.BehaviorExclusionVerifiersOmni = visibles_default,
) -> type[ __.U ]: ...


@__.typx.overload
def dataclass_with_standard_behaviors( # pragma: no cover
    cls: __.AbsentSingleton = __.absent, /, *,
    decorators: __.ClassDecorators[ __.U ] = ( ),
    mutables: __.BehaviorExclusionVerifiersOmni = mutables_default,
    visibles: __.BehaviorExclusionVerifiersOmni = visibles_default,
) -> __.ClassDecoratorFactory[ __.U ]: ...


@__.typx.dataclass_transform( frozen_default = True, kw_only_default = True )
def dataclass_with_standard_behaviors(
    cls: __.Absential[ type[ __.U ] ] = __.absent, /, *,
    decorators: __.ClassDecorators[ __.U ] = ( ),
    mutables: __.BehaviorExclusionVerifiersOmni = mutables_default,
    visibles: __.BehaviorExclusionVerifiersOmni = visibles_default,
) -> type[ __.U ] | __.ClassDecoratorFactory[ __.U ]:
    ''' Decorates dataclass to enforce standard behaviors on instances. '''
    decorate = __.funct.partial(
        __.ccstd.dataclass_with_standard_behaviors,
        attributes_namer = __.calculate_attrname,
        error_class_provider = _iclasses.provide_error_class,
        assigner_core = _iclasses.assign_attribute_if_absent_mutable,
        decorators = decorators,
        mutables = mutables, visibles = visibles )
    if not __.is_absent( cls ): return decorate( )( cls )
    return decorate( )  # No class to decorate; keyword arguments only.


@__.typx.overload
def with_standard_behaviors( # pragma: no cover
    cls: type[ __.U ], /, *,
    decorators: __.ClassDecorators[ __.U ] = ( ),
    mutables: __.BehaviorExclusionVerifiersOmni = mutables_default,
    visibles: __.BehaviorExclusionVerifiersOmni = visibles_default,
) -> type[ __.U ]: ...


@__.typx.overload
def with_standard_behaviors( # pragma: no cover
    cls: __.AbsentSingleton = __.absent, /, *,
    decorators: __.ClassDecorators[ __.U ] = ( ),
    mutables: __.BehaviorExclusionVerifiersOmni = mutables_default,
    visibles: __.BehaviorExclusionVerifiersOmni = visibles_default,
) -> __.ClassDecoratorFactory[ __.U ]: ...


def with_standard_behaviors(
    cls: __.Absential[ type[ __.U ] ] = __.absent, /, *,
    decorators: __.ClassDecorators[ __.U ] = ( ),
    mutables: __.BehaviorExclusionVerifiersOmni = mutables_default,
    visibles: __.BehaviorExclusionVerifiersOmni = visibles_default,
) -> type[ __.U ] | __.ClassDecoratorFactory[ __.U ]:
    ''' Decorates class to enforce standard behaviors on instances. '''
    decorate = __.funct.partial(
        __.ccstd.with_standard_behaviors,
        attributes_namer = __.calculate_attrname,
        error_class_provider = _iclasses.provide_error_class,
        assigner_core = _iclasses.assign_attribute_if_absent_mutable,
        decorators = decorators,
        mutables = mutables, visibles = visibles )
    if not __.is_absent( cls ): return decorate( )( cls )
    return decorate( )  # No class to decorate; keyword arguments only.



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Family of exceptions for package API.

    Provides a hierarchy of exceptions that are raised when accretive behavior
    is violated. The hierarchy is designed to allow both specific and general
    exception handling.
'''


from . import __
from . import classes as _classes


class Omniexception(
    _classes.Object, BaseException,
    instances_visibles = (
        '__cause__', '__context__', __.is_public_identifier ),
):
    ''' Base exceptions for package. '''


class Omnierror( Omniexception, Exception ):
    ''' Base for error exceptions raised by package API. '''


class AttributeImmutability( Omnierror, AttributeError, TypeError ):

    def __init__( self, name: str, target: str ):
        super( ).__init__(
            f"Could not assign or delete attribute {name!r} on {target}." )


class EntryImmutability( Omnierror, TypeError ):

    def __init__( self, indicator: __.cabc.Hashable ) -> None:
        super( ).__init__(
            f"Could not alter or remove existing entry for {indicator!r}." )


class EntryInvalidity( Omnierror, ValueError ):

    def __init__(
        self, indicator: __.cabc.Hashable, value: __.typx.Any
    ) -> None:
        super( ).__init__(
            f"Could not add invalid entry with key, {indicator!r}, "
            f"and value, {value!r}, to dictionary." )


class ErrorProvideFailure( Omnierror, RuntimeError ):

    def __init__( self, name: str, reason: str ):
        super( ).__init__(
            f"Could not provide error class {name!r}. Reason: {reason}" )



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Common imports used throughout the package. '''

# ruff: noqa: F401


import                          abc
import collections.abc as       cabc
import dataclasses as           dcls
import functools as             funct
import                          types

import classcore.standard as    ccstd
import classcore.utilities as   ccutils
import dynadoc as               ddoc
import typing_extensions as     typx
# --- BEGIN: Injected by Copier ---
# --- END: Injected by Copier ---


from absence import AbsentSingleton, Absential, absent, is_absent



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Accretive modules.

    Provides a module type that enforces attribute immutability after
    assignment. This helps ensure that module-level constants remain constant
    and that module interfaces remain stable during runtime.

    The module implementation is derived from :py:class:`types.ModuleType` and
    adds accretive behavior. This makes it particularly useful for:

    * Ensuring constants remain constant
    * Preventing accidental modification of module interfaces

    Also provides a convenience function:

    * ``reclassify_modules``: Converts existing modules to accretive modules.
'''


from . import __
from . import iclasses as _iclasses


ModuleNamespaceDictionary: __.typx.TypeAlias = (
    __.cabc.Mapping[ str, __.typx.Any ] )

ReclassifyModulesModuleArgument: __.typx.TypeAlias = __.typx.Annotated[
    str | __.types.ModuleType | ModuleNamespaceDictionary,
    __.ddoc.Doc( ''' Module, module name, or module namespace. ''' ),
]
ReclassifyModulesRecursiveArgument: __.typx.TypeAlias = __.typx.Annotated[
    bool, __.ddoc.Doc( ''' Recursively reclassify package modules? ''' )
]


class Module(
    __.types.ModuleType,
    metaclass = _iclasses.Class,
    instances_assigner_core = _iclasses.assign_attribute_if_absent_mutable,
):
    ''' Accretive module. '''

    _dynadoc_fragments_ = ( 'module', 'module conceal', 'module accrete' )


def reclassify_modules(
    module: ReclassifyModulesModuleArgument, /, *,
    recursive: ReclassifyModulesRecursiveArgument = False,
) -> None:
    ''' Reclassifies modules to be accretive.

        Can operate on individual modules or entire package hierarchies.

        Only converts modules within the same package to prevent unintended
        modifications to external modules.

        When used with a dictionary, converts any module objects found as
        values if they belong to the same package.

        Has no effect on already-accretive modules.
    '''
    __.ccstd.reclassify_modules(
        module,
        attributes_namer = __.calculate_attrname,
        replacement_class = Module )



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Immutable classes. '''


from . import __


dataclass_core = __.dcls.dataclass( kw_only = True, slots = True )
mutables_default = ( )
visibles_default = ( __.is_public_identifier, )


def assign_attribute_if_absent_mutable( # noqa: PLR0913
    objct: object, /, *,
    ligation: __.AssignerLigation,
    attributes_namer: __.AttributesNamer,
    error_class_provider: __.ErrorClassProvider,
    level: str,
    name: str,
    value: __.typx.Any,
) -> None:
    ''' Assigns attribute if it is absent or mutable, else raises error. '''
    if not hasattr( objct, name ):
        ligation( name, value )
        return
    leveli = 'instance' if level == 'instances' else level
    behaviors_name = attributes_namer( leveli, 'behaviors' )
    behaviors = __.ccutils.getattr0( objct, behaviors_name, frozenset( ) )
    if __.immutability_label not in behaviors:
        ligation( name, value )
        return
    names_name = attributes_namer( level, 'mutables_names' )
    names: __.BehaviorExclusionNamesOmni = (
        getattr( objct, names_name, frozenset( ) ) )
    if names == '*' or name in names:
        ligation( name, value )
        return
    predicates_name = attributes_namer( level, 'mutables_predicates' )
    predicates: __.BehaviorExclusionPredicates = (
        getattr( objct, predicates_name, ( ) ) )
    for predicate in predicates:
        if predicate( name ):
            # TODO? Cache predicate hit.
            ligation( name, value )
            return
    regexes_name = attributes_namer( level, 'mutables_regexes' )
    regexes: __.BehaviorExclusionRegexes = (
        getattr( objct, regexes_name, ( ) ) )
    for regex in regexes:
        if regex.fullmatch( name ):
            # TODO? Cache regex hit.
            ligation( name, value )
            return
    target = __.ccutils.describe_object( objct )
    raise error_class_provider( 'AttributeImmutability' )( name, target )


def provide_error_class( name: str ) -> type[ Exception ]:
    ''' Provides error class for this package. '''
    match name:
        case 'AttributeImmutability':
            from .exceptions import AttributeImmutability as error
        case _:
            from .exceptions import ErrorProvideFailure
            raise ErrorProvideFailure( name, reason = 'Does not exist.' )
    return error


dynadoc_configuration = (
    __.ccstd.dynadoc.produce_dynadoc_configuration(
        introspection = __.dynadoc_introspection_control_on_class,
        table = __.fragments ) )


_class_factory = __.funct.partial(
    __.ccstd.class_factory,
    attributes_namer = __.calculate_attrname,
    dynadoc_configuration = dynadoc_configuration,
    error_class_provider = provide_error_class )


@_class_factory( )
class Class( type ):
    ''' Metaclass for immutable classes. '''

    _dynadoc_fragments_ = (
        'cfc class conceal', 'cfc class protect', 'cfc dynadoc',
        'cfc instance conceal', 'cfc instance protect' )

    def __new__( # Typechecker stub.
        clscls: type[ __.T ],
        name: str,
        bases: tuple[ type, ... ],
        namespace: dict[ str, __.typx.Any ], *,
        decorators: __.ClassDecorators[ __.T ] = ( ),
        **arguments: __.typx.Unpack[ __.ccstd.ClassFactoryExtraArguments ],
    ) -> __.T:
        return super( ).__new__( clscls, name, bases, namespace )


@_class_factory( )
@__.typx.dataclass_transform( frozen_default = True, kw_only_default = True )
class Dataclass( type ):
    ''' Metaclass for immutable dataclasses. '''

    _dynadoc_fragments_ = (
        'cfc produce dataclass',
        'cfc class conceal', 'cfc class protect', 'cfc dynadoc',
        'cfc instance conceal', 'cfc instance protect' )

    def __new__( # Typechecker stub.
        clscls: type[ __.T ],
        name: str,
        bases: tuple[ type, ... ],
        namespace: dict[ str, __.typx.Any ], *,
        decorators: __.ClassDecorators[ __.T ] = ( ),
        **arguments: __.typx.Unpack[ __.ccstd.ClassFactoryExtraArguments ],
    ) -> __.T:
        return super( ).__new__( clscls, name, bases, namespace )


@_class_factory( )
@__.typx.dataclass_transform( kw_only_default = True )
class DataclassMutable( type ):
    ''' Metaclass for immutable dataclasses with mutable instances. '''

    _dynadoc_fragments_ = (
        'cfc produce dataclass',
        'cfc class conceal', 'cfc class protect', 'cfc dynadoc',
        'cfc instance conceal' )

    def __new__( # Typechecker stub.
        clscls: type[ __.T ],
        name: str,
        bases: tuple[ type, ... ],
        namespace: dict[ str, __.typx.Any ], *,
        decorators: __.ClassDecorators[ __.T ] = ( ),
        **arguments: __.typx.Unpack[ __.ccstd.ClassFactoryExtraArguments ],
    ) -> __.T:
        return super( ).__new__( clscls, name, bases, namespace )


@_class_factory( )
class AbstractBaseClass( __.abc.ABCMeta ):
    ''' Metaclass for immutable abstract base classes. '''

    _dynadoc_fragments_ = (
        'cfc produce abstract base class',
        'cfc class conceal', 'cfc class protect', 'cfc dynadoc',
        'cfc instance conceal', 'cfc instance protect' )

    def __new__( # Typechecker stub.
        clscls: type[ __.T ],
        name: str,
        bases: tuple[ type, ... ],
        namespace: dict[ str, __.typx.Any ], *,
        decorators: __.ClassDecorators[ __.T ] = ( ),
        **arguments: __.typx.Unpack[ __.ccstd.ClassFactoryExtraArguments ],
    ) -> __.T: # pragma: no cover
        return super( ).__new__( clscls, name, bases, namespace )


@_class_factory( )
class ProtocolClass( type( __.typx.Protocol ) ):
    ''' Metaclass for immutable protocol classes. '''

    _dynadoc_fragments_ = (
        'cfc produce protocol class',
        'cfc class conceal', 'cfc class protect', 'cfc dynadoc',
        'cfc instance conceal', 'cfc instance protect' )

    def __new__( # Typechecker stub.
        clscls: type[ __.T ],
        name: str,
        bases: tuple[ type, ... ],
        namespace: dict[ str, __.typx.Any ], *,
        decorators: __.ClassDecorators[ __.T ] = ( ),
        **arguments: __.typx.Unpack[ __.ccstd.ClassFactoryExtraArguments ],
    ) -> __.T:
        return super( ).__new__( clscls, name, bases, namespace )


@_class_factory( )
@__.typx.dataclass_transform( frozen_default = True, kw_only_default = True )
class ProtocolDataclass( type( __.typx.Protocol ) ):
    ''' Metaclass for immutable protocol dataclasses. '''

    _dynadoc_fragments_ = (
        'cfc produce protocol class', 'cfc produce dataclass',
        'cfc class conceal', 'cfc class protect', 'cfc dynadoc',
        'cfc instance conceal', 'cfc instance protect' )

    def __new__( # Typechecker stub.
        clscls: type[ __.T ],
        name: str,
        bases: tuple[ type, ... ],
        namespace: dict[ str, __.typx.Any ], *,
        decorators: __.ClassDecorators[ __.T ] = ( ),
        **arguments: __.typx.Unpack[ __.ccstd.ClassFactoryExtraArguments ],
    ) -> __.T:
        return super( ).__new__( clscls, name, bases, namespace )


@_class_factory( )
@__.typx.dataclass_transform( kw_only_default = True )
class ProtocolDataclassMutable( type( __.typx.Protocol ) ):
    ''' Metaclass for immutable protocol dataclasses with mutable instances.
    '''

    _dynadoc_fragments_ = (
        'cfc produce protocol class', 'cfc produce dataclass',
        'cfc class conceal', 'cfc class accrete', 'cfc dynadoc',
        'cfc instance conceal' )

    def __new__( # Typechecker stub.
        clscls: type[ __.T ],
        name: str,
        bases: tuple[ type, ... ],
        namespace: dict[ str, __.typx.Any ], *,
        decorators: __.ClassDecorators[ __.T ] = ( ),
        **arguments: __.typx.Unpack[ __.ccstd.ClassFactoryExtraArguments ],
    ) -> __.T:
        return super( ).__new__( clscls, name, bases, namespace )



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Family of exceptions for package internals. '''


from . import imports as __
from . import nomina as _nomina


class Omniexception(
    __.ccstd.Object, BaseException,
    instances_visibles = (
        '__cause__', '__context__', _nomina.is_public_identifier ),
):
    ''' Base for all exceptions raised internally. '''


class Omnierror( Omniexception, Exception ):
    ''' Base for error exceptions raised internally. '''


class EntryImmutability( Omnierror, TypeError ):
    ''' Attempt to update or remove immutable dictionary entry. '''

    def __init__( self, indicator: __.cabc.Hashable ) -> None:
        super( ).__init__(
            f"Cannot alter or remove existing entry for {indicator!r}." )


class OperationInvalidity( Omnierror, RuntimeError, TypeError ):
    ''' Attempt to perform invalid operation. '''

    def __init__( self, name: str ) -> None:
        super( ).__init__( f"Operation {name!r} is not valid on this object." )



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Docstrings table for reuse across entities. '''


from . import imports as __


fragments: __.cabc.Mapping[ str, str ] = __.types.MappingProxyType( {

    'cfc class accrete':
    ''' By default, accretes class attributes. ''',

    'cfc class conceal':
    ''' By default, conceals non-public class attributes. ''',

    'cfc class protect':
    ''' By default, protects class attributes. ''',

    'cfc dynadoc': ''' Applies Dynadoc decoration to classes. ''',

    'cfc instance accrete':
    ''' Produces classes which can accrete instance attributes. ''',

    'cfc instance conceal':
    ''' Produces classes which can conceal instance attributes. ''',

    'cfc instance protect':
    ''' Produces classes which can protect instance attributes. ''',

    'cfc produce abstract base class':
    ''' Produces abstract base classes compatible with :py:class:`abc.ABCMeta`.
    ''',

    'cfc produce dataclass':
    ''' Produces inheritable dataclasses with keyword-only instantiation. ''',

    'cfc produce protocol class':
    ''' Produces :pep:`544` protocol classes. ''',

    'class accretion':
    ''' By default, class attributes are accretive. ''',

    'class concealment':
    ''' By default, non-public class attributes are invisible. ''',

    'class protection':
    ''' By default, class attributes are immutable. ''',

    'class instance accrete':
    ''' By default, accretes instance attributes. ''',

    'class instance conceal':
    ''' By default, conceals non-public instance attributes. ''',

    'class instance protect':
    ''' By default, protects instance attributes. ''',

    'dataclass':
    ''' Inheritable dataclass with keyword-only instantiation. ''',

    'protocol class':
    ''' Protocol class (:pep:`544`). Nominal and structural subtyping. ''',

    'class dynadoc': ''' Is decorated by Dynadoc. ''',

    'dictionary entries accrete': ''' Accretes dictionary entries. ''',

    'dictionary entries produce':
    ''' Produces default entries on attempt to access absent ones. ''',

    'dictionary entries validate':
    ''' Validates dictionary entries on initialization. ''',

    'module':
    ''' Python module class, derived from :py:class:`types.ModuleType`. ''',

    'module accrete':
    ''' By default, accrets module attributes. ''',

    'module conceal':
    ''' By default, conceals non-public module attributes. ''',

    'namespace':
    ''' Namespace object, modeled after :py:class:`types.SimpleNamespace. ''',

} )
